@@ -3,11 +3,11 @@ from __future__ import absolute_import, division, print_function
 
 import attr
 import six
-from arpeggio import NoMatch, PTNodeVisitor, visit_parse_tree, Terminal
+from arpeggio import NoMatch, PTNodeVisitor, Terminal, visit_parse_tree
 from arpeggio.cleanpeg import ParserPEG
 
-from ._helpers import UNSET
 from . import _segments as segment
+from ._helpers import IMPLICIT_ZERO, UNSET
 
 canonical = r'''
     version = epoch? release pre? post? dev? local? EOF
@@ -19,11 +19,12 @@ canonical = r'''
     pre_post_num = int
     post_tag = "post"
     dev = sep "dev" int
-    local = "+" r'([a-zA-Z0-9]+([-_\.][a-zA-Z0-9]+)*)'
+    local = "+" local_part (sep local_part)*
+    local_part = alpha / int
     sep = dot
     dot = "."
-    int = r'[0-9]+'
-    alpha = r'[a-zA-Z0-9]'
+    int = r'0|[1-9][0-9]*'
+    alpha = r'[0-9]*[a-z][a-z0-9]*'
 '''
 
 permissive = r'''
@@ -38,11 +39,12 @@ permissive = r'''
     post_tag = "post" / "rev" / "r"
     pre_post_num = sep? int
     dev = sep? "dev" int?
-    local = "+" r'([a-zA-Z0-9]+([-_\.][a-zA-Z0-9]+)*)'
+    local = "+" local_part (sep local_part)*
+    local_part = alpha / int
     sep = dot / "-" / "_"
     dot = "."
     int = r'[0-9]+'
-    alpha = r'[a-zA-Z0-9]'
+    alpha = r'[0-9]*[a-z][a-z0-9]*'
 '''
 
 _strict_parser = ParserPEG(canonical, root_rule_name='version', skipws=False)
@@ -50,12 +52,6 @@ _permissive_parser = ParserPEG(
     permissive, root_rule_name='version', skipws=False, ignore_case=True)
 
 
-def unwrap_token(value):
-    if isinstance(value, Token):
-        return value.value
-    return value
-
-
 @attr.s
 class Token(object):
     value = attr.ib()
@@ -106,7 +102,7 @@ class VersionVisitor(PTNodeVisitor):
 
         if sep2 is UNSET:
             sep2 = None
-            num = None
+            num = IMPLICIT_ZERO
 
         assert sep1 is not UNSET
         assert tag is not UNSET
@@ -152,7 +148,7 @@ class VersionVisitor(PTNodeVisitor):
 
         if sep2 is UNSET:
             sep2 = None
-            num = None
+            num = IMPLICIT_ZERO
 
         assert sep1 is not UNSET
         assert tag is not UNSET
@@ -168,7 +164,7 @@ class VersionVisitor(PTNodeVisitor):
         return segment.Post(sep1=UNSET, tag=None, sep2=UNSET, value=children[0])
 
     def visit_dev(self, node, children):
-        num = None
+        num = IMPLICIT_ZERO
         sep = UNSET
 
         for token in children:
@@ -186,7 +182,7 @@ class VersionVisitor(PTNodeVisitor):
         return segment.Dev(value=num, sep=sep)
 
     def visit_local(self, node, children):
-        return segment.Local(''.join(children))
+        return segment.Local(''.join(str(getattr(c, 'value', c)) for c in children))
 
     def visit_int(self, node, children):
         return int(node.value)
@@ -202,9 +198,6 @@ class ParseError(ValueError):
 def parse(version, strict=False):
     parser = _strict_parser if strict else _permissive_parser
 
-    if not strict:
-        version = version.lower()
-
     try:
         tree = parser.parse(version.strip())
     except NoMatch as exc:
