@@ -1,20 +1,22 @@
 # coding: utf-8
 from __future__ import absolute_import, division, print_function
 
+import copy
 import itertools
 import operator
 import re
-from collections import Sequence
 from functools import partial
 
 import attr
 import six
-from attr.validators import in_, instance_of, optional
+from six.moves.collections_abc import Sequence
+from attr.validators import and_, in_, instance_of, optional
 
-from ._helpers import UNSET, Infinity, doc_signature, force_tuple, kwonly_args
-from ._parse import parse
 from . import _segments as segment
-
+from ._helpers import (
+    IMPLICIT_ZERO, UNSET, Infinity, doc_signature, force_lower, force_tuple,
+    kwonly_args, last)
+from ._parse import parse
 
 POST_TAGS = {'post', 'rev', 'r'}
 SEPS = {'.', '-', '_'}
@@ -30,6 +32,42 @@ def unset_or(validator):
     return validate
 
 
+def implicit_or(validator):
+    if isinstance(validator, list):
+        validator = and_(*validator)
+
+    def validate(inst, attr, value):
+        if value == IMPLICIT_ZERO:
+            return
+
+        validator(inst, attr, value)
+    return validate
+
+
+def not_bool(inst, attr, value):
+    if isinstance(value, bool):
+        raise TypeError(
+            "'{name}' must not be a bool (got {value!r})"
+            .format(name=attr.name, value=value)
+        )
+
+
+def is_non_negative(inst, attr, value):
+    if value < 0:
+        raise ValueError(
+            "'{name}' must be non-negative (got {value!r})"
+            .format(name=attr.name, value=value)
+        )
+
+
+def check_by(by, current):
+    if not isinstance(by, int):
+        raise TypeError('by must be an integer')
+
+    if current is None and by < 0:
+        raise ValueError('Cannot bump by negative amount when current value is unset.')
+
+
 validate_post_tag = unset_or(optional(in_(POST_TAGS)))
 validate_pre_tag = optional(in_(PRE_TAGS))
 validate_sep = optional(in_(SEPS))
@@ -39,6 +77,41 @@ is_int = instance_of(six.integer_types)
 is_str = instance_of(six.string_types)
 is_seq = instance_of(Sequence)
 
+# "All numeric components MUST be non-negative integers."
+num_comp = [not_bool, is_int, is_non_negative]
+
+
+def sequence_of(validator, allow_empty=False):
+    if isinstance(validator, list):
+        validator = and_(*validator)
+
+    def validate(inst, attr, value):
+        is_seq(inst, attr, value)
+
+        if not allow_empty and not value:
+            raise ValueError(
+                "'{name}' cannot be empty".format(name=attr.name, value=value)
+            )
+
+        for i, item in enumerate(value):
+            try:
+                validator(inst, attr, item)
+            except (ValueError, TypeError):
+                # now that we know it's invalid, let's re-do the validation
+                # with a better attribute name. Since we have to copy data,
+                # we only do it when we know we have to raise an exception.
+                item_attr = copy.copy(attr)
+                object.__setattr__(item_attr, 'name', '{}[{}]'.format(attr.name, i))
+                try:
+                    validator(inst, item_attr, item)
+                except Exception as exc:
+                    six.raise_from(exc, None)
+                else:
+                    # if we somehow got here, raise original exception
+                    raise
+
+    return validate
+
 
 @attr.s(frozen=True, repr=False, cmp=False)
 class Version(object):
@@ -57,15 +130,15 @@ class Version(object):
         Required to signify a pre-release.
     :type pre_tag: str
 
-    :param pre: `Pre-release`_ number. May be `None` to signify an
+    :param pre: `Pre-release`_ number. May be ``''`` to signify an
         `implicit pre-release number`_.
     :type pre: int
 
-    :param post: `Post-release`_ number. May be `None` to signify an
+    :param post: `Post-release`_ number. May be ``''`` to signify an
         `implicit post release number`_.
     :type post: int
 
-    :param dev: `Developmental release`_ number. May be `None` to signify an
+    :param dev: `Developmental release`_ number. May be ``''`` to signify an
         `implicit development release number`_.
     :type dev: int
 
@@ -104,7 +177,7 @@ class Version(object):
 
         .. doctest::
 
-            >>> v = Version(release=1, post=None)
+            >>> v = Version(release=1, post='')
             >>> str(v)
             '1.post'
             >>> v.post
@@ -116,7 +189,7 @@ class Version(object):
 
         A tuple of integers giving the components of the release segment of
         this :class:`Version` instance; that is, the ``1.2.3`` part of the
-        version number, including trailing zeroes but not including the epoch
+        version number, including trailing zeros but not including the epoch
         or any prerelease/development/postrelease suffixes
 
     .. attribute:: v
@@ -207,14 +280,14 @@ class Version(object):
         #implicit-post-releases
 
     """
-    release = attr.ib(converter=force_tuple, validator=is_seq)
+    release = attr.ib(converter=force_tuple, validator=sequence_of(num_comp))
     v = attr.ib(default=False, validator=is_bool)
-    epoch = attr.ib(default=None, validator=optional(is_int))
+    epoch = attr.ib(default=IMPLICIT_ZERO, validator=implicit_or(num_comp))
     pre_tag = attr.ib(default=None, validator=validate_pre_tag)
-    pre = attr.ib(default=None, validator=optional(is_int))
-    post = attr.ib(default=UNSET, validator=unset_or(optional(is_int)))
-    dev = attr.ib(default=UNSET, validator=unset_or(optional(is_int)))
-    local = attr.ib(default=None, validator=optional(is_str))
+    pre = attr.ib(default=None, validator=implicit_or(optional(num_comp)))
+    post = attr.ib(default=None, validator=implicit_or(optional(num_comp)))
+    dev = attr.ib(default=None, validator=implicit_or(optional(num_comp)))
+    local = attr.ib(default=None, converter=force_lower, validator=optional(is_str))
 
     pre_sep1 = attr.ib(default=None, validator=validate_sep)
     pre_sep2 = attr.ib(default=None, validator=validate_sep)
@@ -230,77 +303,17 @@ class Version(object):
     _key = attr.ib(init=False)
 
     def __attrs_post_init__(self):
-        set = partial(object.__setattr__, self)
+        set_ = partial(object.__setattr__, self)
 
-        if self.epoch is None:
-            set('epoch', 0)
-            set('epoch_implicit', True)
+        if self.epoch == IMPLICIT_ZERO:
+            set_('epoch', 0)
+            set_('epoch_implicit', True)
 
-        if self.pre_tag is not None and self.pre is None:
-            set('pre', 0)
-            set('pre_implicit', True)
+        self._validate_pre(set_)
+        self._validate_post(set_)
+        self._validate_dev(set_)
 
-        if self.pre is not None and self.pre_tag is None:
-            raise ValueError('Must set pre_tag if pre is given.')
-
-        if (self.pre_tag is None and
-                (self.pre_sep1 is not None or self.pre_sep2 is not None)):
-            raise ValueError('Cannot set pre_sep1 or pre_sep2 without pre_tag.')
-
-        if self.post_tag is None:
-            if self.post is UNSET or self.post is None:
-                raise ValueError(
-                    "Implicit post releases (post_tag=None) require a numerical "
-                    "value for 'post' argument.")
-
-            if self.post_sep1 is not UNSET or self.post_sep2 is not UNSET:
-                raise ValueError(
-                    'post_sep1 and post_sep2 cannot be set for implicit post '
-                    'releases (post_tag=None)')
-
-            if self.pre_implicit:
-                raise ValueError(
-                    'post_tag cannot be None with an implicit pre-release '
-                    '(pre=None).')
-
-            set('post_sep1', '-')
-
-        if self.post is not UNSET:
-            if self.post_tag is UNSET:
-                set('post_tag', 'post')
-            if self.post is None:
-                set('post_implicit', True)
-                set('post', 0)
-
-        if self.post_tag is not UNSET and self.post is UNSET:
-            set('post_implicit', True)
-            set('post', 0)
-
-        if self.dev is None:
-            set('dev_implicit', True)
-            set('dev', 0)
-
-        if self.post is UNSET:
-            set('post', None)
-
-        if self.post_tag is UNSET:
-            set('post_tag', None)
-
-        if self.post_sep1 is UNSET:
-            set('post_sep1', None if self.post is None else '.')
-        if self.post_sep2 is UNSET:
-            set('post_sep2', None)
-
-        if self.dev is UNSET:
-            set('dev', None)
-
-        if self.dev_sep is UNSET:
-            set('dev_sep', None if self.dev is None else '.')
-
-        assert self.post_sep1 is not UNSET
-        assert self.post_sep2 is not UNSET
-
-        set('_key', _cmpkey(
+        set_('_key', _cmpkey(
             self.epoch,
             self.release,
             _normalize_pre_tag(self.pre_tag),
@@ -310,6 +323,83 @@ class Version(object):
             self.local,
         ))
 
+    def _validate_pre(self, set_):
+        if self.pre_tag is None:
+            if self.pre is not None:
+                raise ValueError('Must set pre_tag if pre is given.')
+
+            if self.pre_sep1 is not None or self.pre_sep2 is not None:
+                raise ValueError('Cannot set pre_sep1 or pre_sep2 without pre_tag.')
+        else:
+            if self.pre == IMPLICIT_ZERO:
+                set_('pre', 0)
+                set_('pre_implicit', True)
+            elif self.pre is None:
+                raise ValueError('Must set pre if pre_tag is given.')
+
+    def _validate_post(self, set_):
+        got_post_tag = self.post_tag is not UNSET
+        got_post = self.post is not None
+        got_post_sep1 = self.post_sep1 is not UNSET
+        got_post_sep2 = self.post_sep2 is not UNSET
+
+        # post_tag relies on post
+        if got_post_tag and not got_post:
+            raise ValueError('Must set post if post_tag is given.')
+
+        if got_post:
+            if not got_post_tag:
+                # user gets the default for post_tag
+                set_('post_tag', 'post')
+            if self.post == IMPLICIT_ZERO:
+                set_('post_implicit', True)
+                set_('post', 0)
+
+        # Validate parameters for implicit post-release (post_tag=None).
+        # An implicit post-release is e.g. '1-2' (== '1.post2')
+        if self.post_tag is None:
+            if self.post_implicit:
+                raise ValueError(
+                    "Implicit post releases (post_tag=None) require a numerical "
+                    "value for 'post' argument.")
+
+            if got_post_sep1 or got_post_sep2:
+                raise ValueError(
+                    'post_sep1 and post_sep2 cannot be set for implicit post '
+                    'releases (post_tag=None)')
+
+            if self.pre_implicit:
+                raise ValueError(
+                    'post_tag cannot be None with an implicit pre-release '
+                    "(pre='').")
+
+            set_('post_sep1', '-')
+        elif self.post_tag is UNSET:
+            if got_post_sep1 or got_post_sep2:
+                raise ValueError('Cannot set post_sep1 or post_sep2 without post_tag.')
+
+            set_('post_tag', None)
+
+        if not got_post_sep1 and self.post_sep1 is UNSET:
+            set_('post_sep1', None if self.post is None else '.')
+
+        if not got_post_sep2:
+            set_('post_sep2', None)
+
+        assert self.post_sep1 is not UNSET
+        assert self.post_sep2 is not UNSET
+
+    def _validate_dev(self, set_):
+        if self.dev == IMPLICIT_ZERO:
+            set_('dev_implicit', True)
+            set_('dev', 0)
+        elif self.dev is None:
+            if self.dev_sep is not UNSET:
+                raise ValueError('Cannot set dev_sep without dev.')
+
+        if self.dev_sep is UNSET:
+            set_('dev_sep', None if self.dev is None else '.')
+
     @classmethod
     def parse(cls, version, strict=False):
         """
@@ -332,7 +422,7 @@ class Version(object):
             >>> Version.parse('1.dev', strict=True)
             Traceback (most recent call last):
               ...
-            parver._parse.ParseError: Expected int at position (1, 6) => '1.dev*'.
+            parver.ParseError: Expected int at position (1, 6) => '1.dev*'.
         """
         segments = parse(version, strict=strict)
 
@@ -368,12 +458,12 @@ class Version(object):
     def normalize(self):
         return Version(
             release=self.release,
-            epoch=None if self.epoch == 0 else self.epoch,
+            epoch=IMPLICIT_ZERO if self.epoch == 0 else self.epoch,
             pre_tag=_normalize_pre_tag(self.pre_tag),
             pre=self.pre,
-            post=UNSET if self.post is None else self.post,
-            dev=UNSET if self.dev is None else self.dev,
-            local=self.local
+            post=self.post,
+            dev=self.dev,
+            local=_normalize_local(self.local),
         )
 
     def __str__(self):
@@ -463,7 +553,7 @@ class Version(object):
 
         See also: :meth:`clear` and :meth:`replace`.
         """
-        return self.clear(pre=True, post=True, dev=True).replace(local=None)
+        return self.replace(pre=None, post=None, dev=None, local=None)
 
     @property
     def is_prerelease(self):
@@ -511,16 +601,16 @@ class Version(object):
         d = attr.asdict(self, filter=lambda attr, _: attr.init)
 
         if self.epoch_implicit:
-            d['epoch'] = None
+            d['epoch'] = IMPLICIT_ZERO
 
         if self.pre_implicit:
-            d['pre'] = None
+            d['pre'] = IMPLICIT_ZERO
 
         if self.post_implicit:
-            d['post'] = None
+            d['post'] = IMPLICIT_ZERO
 
         if self.dev_implicit:
-            d['dev'] = None
+            d['dev'] = IMPLICIT_ZERO
 
         if self.pre is None:
             del d['pre']
@@ -543,58 +633,11 @@ class Version(object):
 
         return d
 
-    def clear(self, pre=False, post=False, dev=False):
-        """Like :meth:`replace`, but has the ability to **remove** pre-release,
-        post release, and development release segments.
-
-        See also: :meth:`base_version`.
-        """
-        d = self._attrs_as_init()
-
-        if pre:
-            d.pop('pre', None)
-            d.pop('pre_tag', None)
-            d.pop('pre_sep1', None)
-            d.pop('pre_sep2', None)
-
-        if post:
-            d.pop('post', None)
-            d.pop('post_tag', None)
-            d.pop('post_sep1', None)
-            d.pop('post_sep2', None)
-
-        if dev:
-            d.pop('dev', None)
-            d.pop('dev_sep', None)
-
-        return Version(**d)
-
     def replace(self, **kwargs):
         """Return a new :class:`Version` instance with the same attributes,
         except for those given as keyword arguments. Arguments have the same
         meaning as they do when constructing a new :class:`Version` instance
         manually.
-
-        .. warning::
-
-            Be careful! :class:`Version` treats `None` as an implicit zero, so
-            pre-release, post release and development releases cannot be
-            cleared using this method:
-
-            .. doctest::
-
-                >>> Version.parse('1.3.post0').replace(post=None)
-                <Version '1.3.post'>
-                >>> Version.parse('1.3').replace(post=None)
-                <Version '1.3.post'>
-
-            Use :meth:`clear` instead:
-
-            .. doctest::
-
-                >>> Version.parse('1.3.post0').clear(post=True)
-                <Version '1.3'>
-
         """
         d = self._attrs_as_init()
 
@@ -605,6 +648,19 @@ class Version(object):
             d.pop('post_sep1', None)
             d.pop('post_sep2', None)
 
+        if kwargs.get('post', UNSET) is None:
+            kwargs['post_tag'] = UNSET
+            d.pop('post_sep1', None)
+            d.pop('post_sep2', None)
+
+        if kwargs.get('pre', UNSET) is None:
+            kwargs['pre_tag'] = None
+            d.pop('pre_sep1', None)
+            d.pop('pre_sep2', None)
+
+        if kwargs.get('dev', UNSET) is None:
+            d.pop('dev_sep', None)
+
         d.update(kwargs)
         return Version(**d)
 
@@ -635,6 +691,29 @@ class Version(object):
         release = itertools.starmap(new_parts, enumerate(release))
         return self.replace(release=release)
 
+    @doc_signature('(*, by=1)')
+    def bump_epoch(self, **kwargs):
+        """Return a new :class:`Version` instance with the epoch number
+        bumped.
+
+        :param by: How much to bump the number by.
+        :type by: int
+
+        :raises TypeError: `by` is not an integer.
+
+        .. doctest::
+
+            >>> Version.parse('1.4').bump_epoch()
+            <Version '1!1.4'>
+            >>> Version.parse('2!1.4').bump_epoch(by=-1)
+            <Version '1!1.4'>
+        """
+        _, by = kwonly_args(kwargs, (), [('by', 1)])
+        check_by(by, self.epoch)
+
+        epoch = by - 1 if self.epoch is None else self.epoch + by
+        return self.replace(epoch=epoch)
+
     @doc_signature('(*, index)')
     def bump_release(self, **kwargs):
         """Return a new :class:`Version` instance with the release number
@@ -755,17 +834,21 @@ class Version(object):
         _, index, value = kwonly_args(kwargs, ('index', 'value'))
         return self._set_release(index=index, value=value, bump=False)
 
-    def bump_pre(self, tag=None):
+    @doc_signature('(tag=None, *, by=1)')
+    def bump_pre(self, tag=None, **kwargs):
         """Return a new :class:`Version` instance with the pre-release number
         bumped.
 
         :param tag: Pre-release tag. Required if not already set.
         :type tag: str
+        :param by: How much to bump the number by.
+        :type by: int
 
         :raises ValueError: Trying to call ``bump_pre(tag=None)`` on a
             :class:`Version` instance that is not already a pre-release.
         :raises ValueError: Calling the method with a `tag` not equal to the
             current :attr:`post_tag`. See :meth:`replace` instead.
+        :raises TypeError: `by` is not an integer.
 
         .. doctest::
 
@@ -773,8 +856,13 @@ class Version(object):
             <Version '1.4a0'>
             >>> Version.parse('1.4b1').bump_pre()
             <Version '1.4b2'>
+            >>> Version.parse('1.4b1').bump_pre(by=-1)
+            <Version '1.4b0'>
         """
-        pre = 0 if self.pre is None else self.pre + 1
+        _, by = kwonly_args(kwargs, (), [('by', 1)])
+        check_by(by, self.pre)
+
+        pre = by - 1 if self.pre is None else self.pre + by
 
         if self.pre_tag is None:
             if tag is None:
@@ -790,13 +878,18 @@ class Version(object):
 
         return self.replace(pre=pre, pre_tag=tag)
 
-    def bump_post(self, tag=UNSET):
+    @doc_signature('(tag=None, *, by=1)')
+    def bump_post(self, tag=UNSET, **kwargs):
         """Return a new :class:`Version` instance with the post release number
         bumped.
 
         :param tag: Post release tag. Will preserve the current tag by default,
             or use `post` if the instance is not already a post release.
         :type tag: str
+        :param by: How much to bump the number by.
+        :type by: int
+
+        :raises TypeError: `by` is not an integer.
 
         .. doctest::
 
@@ -806,26 +899,77 @@ class Version(object):
             <Version '1.4-1'>
             >>> Version.parse('1.4_post-1').bump_post(tag='rev')
             <Version '1.4_rev-2'>
+            >>> Version.parse('1.4.post2').bump_post(by=-1)
+            <Version '1.4.post1'>
         """
-        post = 0 if self.post is None else self.post + 1
+        _, by = kwonly_args(kwargs, (), [('by', 1)])
+        check_by(by, self.post)
+
+        post = by - 1 if self.post is None else self.post + by
         if tag is UNSET and self.post is not None:
             tag = self.post_tag
         return self.replace(post=post, post_tag=tag)
 
-    def bump_dev(self):
+    @doc_signature('(*, by=1)')
+    def bump_dev(self, **kwargs):
         """Return a new :class:`Version` instance with the development release
         number bumped.
 
+        :param by: How much to bump the number by.
+        :type by: int
+
+        :raises TypeError: `by` is not an integer.
+
         .. doctest::
 
             >>> Version.parse('1.4').bump_dev()
             <Version '1.4.dev0'>
             >>> Version.parse('1.4_dev1').bump_dev()
             <Version '1.4_dev2'>
+            >>> Version.parse('1.4.dev3').bump_dev(by=-1)
+            <Version '1.4.dev2'>
         """
-        dev = 0 if self.dev is None else self.dev + 1
+        _, by = kwonly_args(kwargs, (), [('by', 1)])
+        check_by(by, self.dev)
+
+        dev = by - 1 if self.dev is None else self.dev + by
         return self.replace(dev=dev)
 
+    @doc_signature('(*, min_length=1)')
+    def truncate(self, **kwargs):
+        """Return a new :class:`Version` instance with trailing zeros removed
+        from the release segment.
+
+        :param min_length: Minimum number of parts to keep.
+        :type min_length: int
+
+        .. doctest::
+
+            >>> Version.parse('0.1.0').truncate()
+            <Version '0.1'>
+            >>> Version.parse('1.0.0').truncate(min_length=2)
+            <Version '1.0'>
+            >>> Version.parse('1').truncate(min_length=2)
+            <Version '1.0'>
+        """
+        _, min_length = kwonly_args(kwargs, (), [('min_length', 1)])
+
+        if not isinstance(min_length, int):
+            raise TypeError('min_length must be an integer')
+
+        if min_length < 1:
+            raise ValueError('min_length must be positive')
+
+        release = list(self.release)
+        if len(release) < min_length:
+            release.extend(itertools.repeat(0, min_length - len(release)))
+
+        last_nonzero = max(
+            last((i for i, n in enumerate(release) if n), default=0),
+            min_length - 1,
+        )
+        return self.replace(release=release[:last_nonzero + 1])
+
 
 def _normalize_pre_tag(pre_tag):
     if pre_tag is None:
@@ -841,6 +985,13 @@ def _normalize_pre_tag(pre_tag):
     return pre_tag
 
 
+def _normalize_local(local):
+    if local is None:
+        return None
+
+    return '.'.join(map(str, _parse_local_version(local)))
+
+
 def _cmpkey(epoch, release, pre_tag, pre_num, post, dev, local):
     # When we compare a release version, we want to compare it with all of the
     # trailing zeros removed. So we'll use a reverse the list, drop all the now
